import React, { Suspense, useEffect, useState } from 'react';
import {
  Edge,
  Node,
  useAppDispatch,
  useGraphQueryResult,
  useGraphQueryResultMeta,
  useML,
  useQuerybuilderGraph,
  useSchemaGraph,
  useSessionCache,
  useVisualization,
} from '@graphpolaris/shared/lib/data-access';
import VisualizationTabBar from './VisualizationTabBar';
import { Recommender, NoData, Querying } from '../views';
import { resultSetFocus, resultSetSelection, unSelect } from '../../data-access/store/interactionSlice';
import { updateVisualization, addVisualization } from '../../data-access/store/visualizationSlice';
import { VisualizationPropTypes, VISComponentType } from '../common';

type PromiseFunc = () => Promise<{ default: VISComponentType<any> }>;
export const Visualizations: Record<string, PromiseFunc> = {
  TableVis: () => import('../visualizations/tablevis/tablevis'),
  PaohVis: () => import('../visualizations/paohvis/paohvis'),
  RawJSONVis: () => import('../visualizations/rawjsonvis/rawjsonvis'),
  NodeLinkVis: () => import('../visualizations/nodelinkvis/nodelinkvis'),
  MatrixVis: () => import('../visualizations/matrixvis/matrixvis'),
  SemanticSubstratesVis: () => import('../visualizations/semanticsubstratesvis/semanticsubstratesvis'),
  MapVis: () => import('../visualizations/mapvis/mapvis'),
};
export const VISUALIZATION_TYPES: string[] = Object.keys(Visualizations);

export const VisualizationPanel = ({ fullSize }: { fullSize: () => void }) => {
  const query = useQuerybuilderGraph();
  const graphQueryResult = useGraphQueryResult();
  const dispatch = useAppDispatch();
  const { activeVisualizationIndex, openVisualizationArray } = useVisualization();
  const ml = useML();
  const schema = useSchemaGraph();
  const graphMetadata = useGraphQueryResultMeta();
  const [viz, setViz] = useState<{ component: React.FC<VisualizationPropTypes>; id: string; exportImage: () => void } | undefined>(
    undefined,
  );

  useEffect(() => {
    if (openVisualizationArray.length > 0) {
      loadVisualization();
    }
  }, [activeVisualizationIndex, openVisualizationArray.length]);

  const loadVisualization = async (add = false) => {
    if (activeVisualizationIndex === -1) {
      setViz(undefined);
      return;
    }

    const activeVisualization = openVisualizationArray[activeVisualizationIndex];

    if (!activeVisualization) {
      setViz(undefined);
      return;
    }

    const componentModule = await Visualizations[activeVisualization.id]();
    const component = componentModule.default;

    if (add) {
      dispatch(addVisualization(component.settings));
    }
    // Extract exportImage function or default to no-op
    const exportImage = component.exportImage || (() => {});

    setViz({ component: component.component, id: activeVisualization.id, exportImage });
  };

  const handleSelect = (selection?: { nodes?: Node[]; edges?: Edge[] }) => {
    if (selection?.nodes && selection.nodes.length > 0) dispatch(resultSetSelection(selection.nodes));
    else dispatch(unSelect());
  };

  const updateSettings = (newSettings: Record<string, any>) => {
    if (activeVisualizationIndex) {
      const updatedSettings = { ...openVisualizationArray[activeVisualizationIndex], ...newSettings };
      dispatch(updateVisualization({ id: activeVisualizationIndex, settings: updatedSettings }));
    }
  };
  const exportImage = viz?.exportImage || (() => {});
  return (
    <div
      className="relative pt-7 vis-panel h-full w-full flex flex-col border bg-light"
      onMouseDownCapture={() => dispatch(resultSetFocus({ focusType: 'visualization' }))}
    >
      <div className="grow overflow-y-auto" style={graphQueryResult.nodes.length === 0 ? { overflow: 'hidden' } : {}}>
        {graphQueryResult.queryingBackend ? (
          <Querying />
        ) : graphQueryResult.nodes.length === 0 ? (
          <NoData dataAvailable={query.nodes.length > 0} />
        ) : openVisualizationArray.length === 0 ? (
          <Recommender />
        ) : (
          <div className="w-full h-full flex">
            <Suspense fallback={<div>Loading...</div>}>
              {!!viz &&
                activeVisualizationIndex !== -1 &&
                openVisualizationArray?.[activeVisualizationIndex] &&
                viz.id === openVisualizationArray[activeVisualizationIndex].id &&
                graphMetadata && (
                  <viz.component
                    data={graphQueryResult}
                    schema={schema}
                    ml={ml}
                    settings={openVisualizationArray[activeVisualizationIndex]}
                    dispatch={dispatch}
                    handleSelect={handleSelect}
                    graphMetadata={graphMetadata}
                    updateSettings={updateSettings}
                    handleHover={() => {}}
                  />
                )}
            </Suspense>
          </div>
        )}
      </div>
      <VisualizationTabBar fullSize={fullSize} exportImage={exportImage} />
    </div>
  );
};
