import { PaohVisProps } from '../paohvis';
import { PaohvisData, HyperEdgeRange, HyperEdgeI } from '../types';
import AttributeFilterUsecase, { getIds } from './AttributesFilterUseCase';
import { countRepetition } from './utils';
import { Edge, GraphQueryResult, Node } from '@graphpolaris/shared/lib/data-access';

type AugmentedNode = Node & {
  reduced_ids: string[];
};

function parseNodes(nodes: Node[], yAxisNodeType: string, xAxisNodeType: string): string[] {
  //const rowNodes = filterRowNodes(nodes, hyperEdgeDegree, yAxisNodeType);
  const rowNodes = nodes;
  const rowLabels = getIds(rowNodes);

  //make dictionary for finding the index of a row
  const yNodesIndexDict: Record<string, number> = {};
  let yNodeIndexCounter = 0;
  for (let i = 0; i < rowNodes.length; i++) {
    yNodesIndexDict[rowNodes[i]._id] = yNodeIndexCounter;
    yNodeIndexCounter++;
  }

  return rowLabels;
}

function parseHyperEdgeRanges(
  nodesColumn: AugmentedNode[],
  nodesRow: AugmentedNode[],
  filteredData: { nodes: Node[]; edges: Edge[] },
  rowInfo: string[],
  nodeIdIndex: { [id: string]: number },
  columnNode: string,
  relationTo: string,
): [HyperEdgeRange[], { [key: string]: number }] {
  if (nodesColumn.length == 0 || nodesRow.length == 0) return [[], {}];

  let resultHyperEdgeRanges: HyperEdgeRange[] = nodesColumn.map((element) => {
    const rangeText: any = +element._id;
    const _id = element._id;
    const degree = 0;
    const hyperEdges: HyperEdgeI = {
      indices: [],
    };

    return {
      rangeText,
      hyperEdges,
      _id,
      reduced_ids: element.reduced_ids,
      degree,
    };
  });

  const processedPairs = new Set<string>(); // To avoid processing pairs

  const toOrFrom = relationTo == columnNode ? 'to' : 'from';
  const toOrFromOpposite = toOrFrom == 'to' ? 'from' : 'to';

  //console.log(relationTo, columnNode, relationTo == columnNode);

  // loop through all edges - and creates hyperedge structure (column order/x)
  for (let i = 0; i < filteredData.edges.length; i++) {
    const edge = filteredData.edges[i];

    const pairString = `${edge.from}-${edge.to}`;
    if (!processedPairs.has(pairString)) {
      const hyperEdgeRange = resultHyperEdgeRanges.find((range) => range._id === edge[toOrFrom])?.hyperEdges;
      hyperEdgeRange?.indices.push(nodeIdIndex[edge[toOrFromOpposite]]);
      processedPairs.add(pairString);
    }
  }

  // Sequential order in indices and count degree ( remove undefined = nodes without connection )
  resultHyperEdgeRanges.forEach((element) => {
    element.degree = element.hyperEdges.indices.filter((num) => {
      //console.log(num, typeof num);
      if (typeof num !== 'number') {
        //console.log('Filtered out:', num);
      }
      return typeof num === 'number';
    }).length;
  });

  resultHyperEdgeRanges.forEach((element) => {
    element.hyperEdges.indices.sort((a, b) => a - b);
  });

  // Count nodes degree from structure
  const repetitionCounts: { [key: number]: number } = {};

  for (let indexX = 0; indexX < resultHyperEdgeRanges.length; indexX++) {
    countRepetition(resultHyperEdgeRanges[indexX].hyperEdges.indices, repetitionCounts);
  }

  const repetitionCountLabel: { [key: string]: number } = {};
  // get nodes ids
  rowInfo.forEach((arrayElement, index) => {
    repetitionCountLabel[arrayElement] = repetitionCounts[index] === undefined ? 0 : repetitionCounts[index];
  });

  resultHyperEdgeRanges.sort((a, b) => a.rangeText - b.rangeText);

  return [resultHyperEdgeRanges, repetitionCountLabel];
}

export function parseQueryResult(
  queryResult: GraphQueryResult,
  configuration: PaohVisProps,
  relationTo: string,
  mergeData: boolean,
): PaohvisData {
  const skipReduceRow = configuration.attributeRowShow.includes('_id');
  const skipReduceColumn = configuration.attributeColumnShow.includes('_id');
  const replaceNodeIds: Record<string, string> = {};
  const reducedReplaceNodeIds: Record<string, string[]> = {};
  const hashAttributesToIdMap: Record<string, string[]> = {};
  let nodesRow: AugmentedNode[] = queryResult.nodes
    .filter((node) => {
      // some dataset do not have label field
      let labelNode = '';
      if (node.label !== undefined) {
        labelNode = node.label;
      } else {
        const idParts = node._id.split('/');
        labelNode = idParts[0];
      }
      return labelNode === configuration.rowNode;
    })
    .map((node) => {
      const attributes = skipReduceRow
        ? node.attributes
        : Object.fromEntries(
            Object.keys(node.attributes)
              .filter((k) => new Set(configuration.attributeRowShow).has(k))
              .map((k) => [k, node.attributes[k]]),
          );

      const hash = JSON.stringify(attributes);
      if (hash in hashAttributesToIdMap) {
        hashAttributesToIdMap[hash].push(node._id);
      } else {
        hashAttributesToIdMap[hash] = [node._id];
      }
      return { ...node, attributes: attributes, reduced_ids: [node._id] };
    });

  let nodesColumn: AugmentedNode[] = queryResult.nodes
    .filter((node) => {
      let labelNode = '';
      if (node.label !== undefined) {
        labelNode = node.label;
      } else {
        const idParts = node._id.split('/');
        labelNode = idParts[0];
      }
      return labelNode === configuration.columnNode;
    })
    .map((node) => {
      const attributes = skipReduceColumn
        ? node.attributes
        : Object.fromEntries(
            Object.keys(node.attributes)
              .filter((k) => new Set(configuration.attributeColumnShow).has(k))
              .map((k) => [k, node.attributes[k]]),
          );
      const hash = JSON.stringify(attributes);
      if (hash in hashAttributesToIdMap) {
        hashAttributesToIdMap[hash].push(node._id);
      } else {
        hashAttributesToIdMap[hash] = [node._id];
      }
      return { ...node, attributes: attributes, reduced_ids: [node._id] };
    });

  // Make id->id replace map
  Object.entries(hashAttributesToIdMap).forEach(([k, v]) => {
    if (v.length <= 1) return;

    v.slice(1).forEach((id) => {
      replaceNodeIds[id] = v[0];
      if (!(id in reducedReplaceNodeIds)) reducedReplaceNodeIds[id] = [v[0]];
      else reducedReplaceNodeIds[id].push(v[0]);
    });
  });

  const toRemoveIds = new Set(Object.keys(replaceNodeIds));
  if (mergeData) {
    nodesRow = nodesRow
      .filter((node) => !toRemoveIds.has(node._id))
      .map((node) => ({ ...node, reduced_ids: [node._id, ...(reducedReplaceNodeIds?.[node._id] || [])] }));
    nodesColumn = nodesColumn
      .filter((node) => !toRemoveIds.has(node._id))
      .map((node) => ({ ...node, reduced_ids: [node._id, ...(reducedReplaceNodeIds?.[node._id] || [])] }));
  }

  const augmentedNodes = [...nodesRow, ...nodesColumn];
  const augmentedEdges = queryResult.edges.map((edge) => {
    return {
      _id: edge._id,
      from: replaceNodeIds[edge.from] || edge.from,
      to: replaceNodeIds[edge.to] || edge.to,
      label: edge.label,
      attributes: edge.attributes,
    };
  });

  const nodeIdIndex: { [id: string]: number } = nodesRow.reduce((acc, node, index) => ({ ...acc, [node._id]: index }), {});

  const paohvisFilters = { nodeFilters: [], edgeFilters: [] }; // fill it with something
  const filteredData = AttributeFilterUsecase.applyFilters(augmentedNodes, augmentedEdges, paohvisFilters);

  //parse nodes
  const rowInfo: string[] = parseNodes(nodesRow, configuration.rowNode, configuration.columnNode);

  let nodeListAttr: any[] = [];
  nodeListAttr = nodesRow.map((node) => node._id);

  const [resultHyperEdgeRanges, rowsDegree] = parseHyperEdgeRanges(
    nodesColumn,
    nodesRow,
    filteredData,
    rowInfo,
    nodeIdIndex,
    configuration.columnNode,
    relationTo,
  );

  return {
    rowLabels: nodeListAttr,
    hyperEdgeRanges: resultHyperEdgeRanges,
    rowDegrees: rowsDegree,
    nodes: augmentedNodes,
    edges: augmentedEdges,
  };
}
